import { defineQuery, hasComponent } from 'bitecs'
import { query as q } from './query-parser/index.js'
import { parseAll } from './parser.js'
import { curry, of, is, reduce, flip, concat, map, when, assoc, mergeLeft, isNil, path, prop, ifElse, diverge, pipe, always, assocPath, prepend, orDefault, useWith, applyTo, tap } from './fn.js'
import { Relationship } from './query-parser/types.js'

const prepareQuery = (query, component) => {
  const { match, returnValues } = query
  const relationship = is(Relationship, match)

  if (relationship) {
    const { from: f, to: t, edge: e } = match
    const [from, to, edge] = map(q => prepareQuery({ match: q, returnValues }, component), [f, t, e])

    return { from, to, edge, relationship, schema: query }
  } else {
    const label = path('label.value.value', match)
    const name = path('name.value.value', match)
    const type = component[label]

    return { name, label, type, relationship, query: defineQuery([type]), schema: query }
  }
}

const matches = (world, type, query, set = new Set()) => entity => {
  if (set.has(entity)) {
    return true
  } else if (hasComponent(world, type, entity) && query.includes(entity)) {
    set.add(entity)
    return true
  } else {
    return false
  }
}

const queryRelationship = (from, edge, to, world) => {
  const matchesFrom = matches(world, from.type, from.results)
  const matchesTo = matches(world, to.type, to.results)

  const Edge = edge.type

  const result = []
  for (let i = 0; i < edge.results.length; i++) {
    const eid = edge.results[i]

    if (matchesFrom(Edge.from[eid]) && matchesTo(Edge.to[eid])) {
      result.push(eid)
    }
  }

  return result
}

const assembleQuery = curry((query, entities) => map(entity => of({ entity, query }), entities))

const executeQuery = curry((query, world) => {
  if (!query.relationship) {
    return assembleQuery(query, query.query(world))
  }

  const { from, to: _to, edge } = query
  const to = when(prop('relationship'), prop('from'), _to)

  const edges = queryRelationship(
    { type: from.type, results: from.query(world) },
    { type: edge.type, results: edge.query(world) },
    { type: to.type, results: to.query(world) },
    world
  )

  if (_to.relationship) {
    return reduce((acc, eid) => {
      const next = assocPath(
        'from.query',
        always(of(edge.type.to[eid])),
        _to
      )

      return pipe(
        executeQuery(next),
        map(prepend(eid)),
        map(([l, r]) => ([{ entity: l, query }, r])),
        concat(acc)
      )(world)
    }, [], edges)
  } else {
    return assembleQuery(query, edges)
  }
})

const returnValues = pipe(path('query.schema.returnValues.values'), map(prop('value')))
const nodeName = path('query.schema.match.name.value.value')
const includes = curry((val, arr) => !isNil(val) && arr.includes(val))

const maybeAssoc = curry((pred, key, value, obj) =>
  when(pred, assoc(key, value), obj)
)

const resolveNode = curry((returns, name, entity, obj) =>
  maybeAssoc(always(includes(name, returns)), name, entity, obj)
)

const resolveRelationship = edges => {
  const { entity, query } = edges
  const Edge = query.edge.type
  const returns = returnValues(edges)

  const queryProp = flip(prop)(query)
  const [from, edge, to] = map(queryProp, ['from', 'edge', 'to'])
  const [fn, en] = [from.name, edge.name]
  const tn = ifElse(prop('relationship'),
    path('from.name'),
    prop('name'),
    to
  )

  const resolve = resolveNode(returns)
  return pipe(
    resolve(fn, Edge.from[entity]),
    resolve(en, entity),
    resolve(tn, Edge.to[entity]),
  )({})
}

const resolveReturns = map(
  pipe(
    map(ifElse(
      path('query.relationship'),
      resolveRelationship,
      diverge(resolveNode, [returnValues, nodeName, prop('entity'), always({})])
    )),
    reduce(mergeLeft, {})
  )
)

export const query = curry((input, { world, component }) =>
  map(({ use, ...rest }) =>
    pipe(
      prop('value'),
      orDefault('default'),
      flip(prop)(world),
      executeQuery(prepareQuery(rest, component)),
      resolveReturns
    )(use),
    parseAll(q, input)
  )
)

