export class NotImplementedError extends Error {
  constructor(msg) {
    super(msg)
  }
}

export class Iterator {
  static handledTypes = new Map()

  static from(value) {
    if (!value) { return }
    const types = Iterator.handledTypes
    const ctr = value.constructor || value.prototype.constructor

    if (types.has(ctr)) {
      const handler = types.get(ctr)
      return new handler(value)
    } else {
      throw new NotImplementedError(`${value} is not an Iterator`)
    }
  }

  next() {
    throw new NotImplementedError('Iterator::next not implemented')
  }

  peek() {
    throw new NotImplementedError('Iterator::peek not implemented')
  }

  clone() {
    throw new NotImplementedError('Iterator::clone not implemented')
  }

  done() {
    return !!this.peek()
  }

  take(n) {
    let accumulator = []

    for (let i = 0; i < n; i++) {
      accumulator.push(this.next())
    }

    return accumulator
  }

  drop(n) {
    this.take(n)
    return this
  }
}

class ArrayIterator extends Iterator {
  static { Iterator.handledTypes.set(Array, ArrayIterator) }
  #index = -1
  #value

  constructor(value, index = -1) {
    super()

    if (!Array.isArray) {
      throw new Error(`${value} is not an Array`)
    }

    this.#index = index
    this.#value = value
  }

  next() {
    if (this.#index <= this.#value.length) {
      this.#index += 1
      return this.#value[this.#index]
    }
  }

  peek() {
    return this.#value[this.#index]
  }

  clone() {
    return new ArrayIterator(this.#value, this.#index)
  }
}

class StringIterator extends ArrayIterator {
  static { Iterator.handledTypes.set(String, StringIterator) }

  constructor(value) {
    super(value.split(''))
  }
}

export class Stream extends Iterator {
  constructor(value) {
    super()
    if (!value.next || !typeof value.next === 'function') {
      this.iterator = Iterator.from(value)

      if (!value) {
        throw new Error(`${value} is not an iterator`)
      }
    } else {
      this.iterator = value
    }
  }

  next() {
    return this.iterator.next()
  }

  peek() {
    return this.iterator.peek()
  }

  clone() {
    return new Stream(this.iterator.clone())
  }
}

